/*

  Simple C-like dynamic vectors/matrices

  use doble, no typedef no template

*/

#include <cstddef>
#include <iostream>

// if C++ < C++11 define nullptr
#if __cplusplus <= 199711L
  #include <cstdlib>
  #ifndef nullptr
    #include <cstddef>
    #define nullptr NULL
  #endif
#endif

/*
__     __        
\ \   / /__  ___ 
 \ \ / / _ \/ __|
  \ V /  __/ (__ 
   \_/ \___|\___|
*/
                 
using namespace std ;

// define a vector as a struct with the size of the vector
// and a pointer to its values

class Vec {
  int      _size ;
  double * values ;

public:
  
  Vec() ; // default constructor
  Vec( int _size ) ; // another constructor

  ~Vec() ; // destructor

  Vec( Vec const & ) ; // copy constructor
  Vec const & operator = ( Vec const & ) ; // copy operator

  int size() const { return _size ; }
  
  // methods
  void allocate( int size ) ;
  void fill( double val ) ;
  void fill_random() ;

  // magic methods, redefine operators!
  double
  operator () // name of the operator
  ( int i ) const // const is used to inform the compiler that operator do not modifty contents
  { return values[i] ; }   

  double & // return the reference (i.e. the address) of values[i]
  operator () ( int i )
  { return values[i] ; }   

} ;

/*
 __  __       _   
|  \/  | __ _| |_ 
| |\/| |/ _` | __|
| |  | | (_| | |_ 
|_|  |_|\__,_|\__|
*/

class Mat {
  int       _nrow, _ncol ;
  double  * memory_block ; // allocated memory for the matrix
  double ** values ; // pointer to the rows of the matrix

public:

  Mat() ; // default constructor
  Mat( int _nr, int _nc ) ; // another constructor

  Mat( Mat const & ) ; // copy constructor
  Mat const & operator = ( Mat const & ) ; // copy operator

  ~Mat() ; // destructor

  int ncol() const { return _ncol ; }
  int nrow() const { return _nrow ; }
  
  void allocate( int nrow, int ncol ) ;
  void fill_random( ) ;
  void fill( double val ) ;

  // magic methods, redefine operators!
  double
  operator () // name of the operator
  ( int i, int j ) const // const is used to inform the compiler that operator do not modifty contents
  { return values[i][j] ; }

  double & // return the reference (i.e. the address) of values[i]
  operator () ( int i, int j )
  { return values[i][j] ; }

} ;

/*
 ___    _____  
|_ _|  / / _ \ 
 | |  / / | | |
 | | / /| |_| |
|___/_/  \___/ 

*/

ostream &
operator << ( ostream & stream, Vec const & vec ) ;

ostream &
operator << ( ostream & stream, Mat const & mat ) ;

/*
    _    _            _               
   / \  | | __ _  ___| |__  _ __ __ _ 
  / _ \ | |/ _` |/ _ \ '_ \| '__/ _` |
 / ___ \| | (_| |  __/ |_) | | | (_| |
/_/   \_\_|\__, |\___|_.__/|_|  \__,_|
           |___/                      
*/

Vec operator + ( Vec const & a, Vec const & b ) ;
Vec operator - ( Vec const & a, Vec const & b ) ;
Vec operator * ( Vec const & a, Vec const & b ) ;
Vec operator / ( Vec const & a, Vec const & b ) ;

Vec operator * ( double a, Vec const & b ) ;
Vec operator * ( Vec const & a, double b ) ;

Vec operator / ( Vec const & a, double b ) ;
Vec operator ^ ( Vec const & a, double b ) ;

double dot( Vec const & a, Vec const & b ) ;
double norm( Vec const & a ) ;

// ------------

Mat operator + ( Mat const & A, Mat const & B ) ;
Mat operator - ( Mat const & A, Mat const & B ) ;
Mat operator * ( Mat const & A, Mat const & B ) ;

Mat operator * ( double a, Mat const & B ) ;
Mat operator * ( Mat const & A, double b ) ;

Mat operator / ( Mat const & A, double b ) ;

// ------------
Vec operator * ( Mat const & A, Vec const & b ) ;
